/*
 * W2000A Screenshot Utility v0.92b
 * $Rev$
 */

#define	REVMAJ	0
#define	REVMIN	92

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <strings.h>
#ifndef WIN32
#include <fcntl.h>
#include <err.h>
#include <termios.h>
#include <time.h>
#else
#include <windows.h>
#include "ComTools.h"
#endif

static void usage(void);
static void set_serial(int);
static int open_serial(char *);
static int rdchar(int);
static int wrchar(int, char);
static void receive_trace(int, int, const char *, char);	//KB
static void receive_screen(int, int);
static void retrieve(int, int, int);
static void combine_screen(int, int);
static time_t gettime(void);

/*
 * XXX: Please note:
 *      This is _not_ 100% identical to the actual
 *      output seen on the DSO.  For example:
 *      The math channel is drawn _over_
 *      the browse bar.  Put the Math channel
 *      behind UI in the planeorder to witness
 *      the actual output.
 */
static const int planeorder[] = {
	7, 8, 9, 10, 		/* Channels 1-4 */
				/* XXX: Red Arrow and Stop-Sign in UI on 10/Channel4? */
	11,			/* Math Channel */
	12, 13,			/* Cursors and Trigger Mark */
	5, 4, 6, 2, 3,		/* UI */
//	12, 13, 14,		/* Not yet determined */
	1,			/* Grid */
	0
};

static const int planecolours[][3] = {
			/* R     G     B */
/* UI_Plane1 */		{ 0xFB, 0xFB, 0xFB }, /* dark white */
/* UI_Plane2 */		{ 0x00, 0x00, 0x00 }, /* black */
/* UI_Plane4 */		{ 0xE6, 0xE6, 0x83 }, /* beige-ish yellow */
/* UI_Plane3 */		{ 0xFF, 0xFD, 0xEE }, /* yellow tinted white */
/* UI_Plane5 */		{ 0x7E, 0x7E, 0x7E }, /* gray */
/* Channel_Plane1 */	{ 0xFF, 0xFF, 0x00 }, /* yellow */
/* Channel_Plane2 */	{ 0x00, 0xFF, 0x00 }, /* light green */
/* Channel_Plane3 */	{ 0x32, 0xAB, 0xFF }, /* light blue */
/* Channel_Plane4 */	{ 0xFF, 0x00, 0x00 }, /* red */
/* Channel_Math_Plane */{ 0xFF, 0x96, 0xF6 }, /* pink */
// /* Memory_Plane1 */	{ 0x00, 0x00, 0x00 }, /* n.y.d. */
// /* Memory_Plane2 */	{ 0x00, 0x00, 0x00 }, /* n.y.d. */
// /* Memory_Plane3 */	{ 0x00, 0x00, 0x00 }, /* n.y.d. */
/* Marker_Plane1 */	{ 0xFF, 0x67, 0x00 }, /* orange-red */
/* Marker_Plane2 */	{ 0xFF, 0x67, 0x00 }, /* orange-red */
/* Grid_Plane */	{ 0x3F, 0x3F, 0x3F }, /* dark gray */
/* Background */	{ 0x00, 0x00, 0x00 }, /* black; don't remove this! */
/* B/W Screenshot */	{ 0xFF, 0xFF, 0xFF }, /* white; don't remove this! */
};
static const unsigned int maxplanes=sizeof(planecolours)/3/sizeof(int);

static unsigned char buffer[640*480/2];					//big enough for color and BW
static unsigned char *bufptr, *bufend;					
static char fnamprefix[256];
static int fnum;
static unsigned int rdplanes;
static time_t last_t, now_t;
static int iflag;
static char tracetype[16];

#ifdef WIN32
static void
err(int e, const char *fmt, ...)
{

	/*
	 * Don't even know if win32 has sth. like vsnprintf(),
	 * so we don't care.
	 */
	(void)fprintf(stderr, "Error: %s, exiting.\n", fmt);
	exit(e);
}

static void
errx(int e, const char *fmt, ...)
{

	err(e, fmt);
}
#endif

static void
usage(void)
{

	(void)fprintf(stderr,
			"Usage:\n"
#ifdef WIN32
			"w2000a-screenshot.exe [-v] [-h] [-c n] [-f prefix] [-n num] [-b] [-smdr] [-o] [-t type]\n"
#else
			"w2000a-screenshot [-v] [-h] [-c n] [-f prefix] [-n num] [-b] [-smdr] [-o] [-t type]\n"
#endif
			" -h\t\tthis help\n"
			" -v\t\tVersion information\n"
#ifdef WIN32
			" -c n\t\tCOM port (0 < n < 10)\n"
#else
			" -c n\t\tSerial device, e.g. /dev/ttyS0\n"
#endif
			" -f prefix\toutput file\n"
			" -n num\t\tdefine start of file numbering\n"
			" -b\t\twrite as BMP file, not P(P|G)M\n"
			" -s\t\tinitiate standard screenshot (colour dump)\n"
			" -m\t\tinitiate monochrome dump\n"
			" -d\t\tinitiate trace dump from filtered data\n"
			" -r\t\tinitiate trace dump from UNfiltered data\n"
			" -o\t\tOne-Shot, do not wait for more dumps (set on [-smd])\n"
			" -i\t\tInvert colours (i.e. for monochromatic screenshot)\n"
			" -t type\tspecify trace file type, default: csv, valid: csv, ascii\n"
			"\n"
#ifdef WIN32
			"Input is read from specified COM port,\n"
			"the default is COM1.\n"
#else
			"Input is read from serial port, diagnostics and\n"
			"status messages are directed to stderr.\n"
			"Default is /dev/ttyS0\n"
#endif
			"\n"
			"If P(P|G)M (default) output is in use, planes\n"
			"are written to \"[prefix]-[num]_dd.pgm\", with\n"
			"\"dd\" as the two digit plane sequence number.\n"
			"Prefix defaults to \"screenshot\".\n"
			"\n"
			"Planes are coloured and combined in \"[prefix]-[num].[suf]\",\n"
			"with suffix ppm or bmp.\n"
			"\n"
			"Trace data is written to [prefix]-[num].[typesuf].\n");
	exit(1);
}

static void
set_serial(int com)
{

#ifdef WIN32
	(void)com; /* unused */
#else
	struct termios termios;

	if (tcgetattr(com, &termios) == -1)
		err(1, "tcgetattr() for serial port failed");
	if (cfsetspeed(&termios,B115200) == -1)
		err(1, "cfsetspeed() failed, unable to set desired speed (115200 Baud)");
	cfmakeraw(&termios);
	termios.c_cc[VTIME]=150;
	termios.c_cc[VMIN]=1;
	if (tcsetattr(com, TCSANOW, &termios) == -1)
		err(1, "tcsetattr() failed, unable to set line attributes");
#endif
}

static int
open_serial(char *device)
{
	int com;

#ifdef WIN32
	if (device == NULL)
		com = 1;
	else
		com = *device - '0';
	if (!ComOpen(com - 1, 115200, P_NONE, S_1BIT, D_8BIT))
		err(1, "Unable to open COM port");
#else
	if (device == NULL)
		device = (char *)"/dev/ttyS0";
	if ((com = open(device, O_RDWR)) == -1)
		err(1, "open(%s) failed", device);
	if (isatty(com)) set_serial(com);
#endif

	return com;
}

static int
rdchar(int com)
{
#ifdef WIN32
	while (ComGetReadCount(com - 1) < 1)
		Sleep(1);
	return ComRead(com - 1);
#else
	unsigned char c;

	if (read(com, &c, 1) > 0)
		return c;
	return -1;
#endif
}

static int
wrchar(int com, char c)
{

#ifdef WIN32
	return ComWrite(com - 1, c);
#else
	return write(com, &c, 1);
#endif
}

static void
receive_trace(int com, int type, const char *ftype, char format)
{
	FILE *fp;
	unsigned char cdata[4][16384];
	char fnam[256];
	int incnt=0,r, chan, idx;

	(void)type; /* unused */

	if (type==2) {
		(void)snprintf(fnam, sizeof(fnam), "%s-%.4d.%s", fnamprefix, fnum, ftype);
		(void)fprintf(stderr, "  - Receiving trace data (file: %s)\n", fnam); }
	if (type==3) {
		(void)snprintf(fnam, sizeof(fnam), "%s-%.4d-raw.%s", fnamprefix, fnum, ftype);
		(void)fprintf(stderr, "  - Receiving (raw) trace data (file: %s)\n", fnam); }
	if ((fp = fopen(fnam, "w+")) == NULL)
		err(1, "fopen(%s, w+) failed", fnam);
	fnum++;

/*********************************************************************
Changed due to necessary changes in FW
Now all data for each channel are sent in order.
*********************************************************************/
	last_t = gettime();

	for (chan=0;chan<4;chan++) {
		for (idx=0;idx<16384;idx++) {
			r = rdchar(com);
			cdata[chan][idx]=(unsigned char)r;
			incnt++;
			if (incnt % 100 == 0)
				(void)fprintf(stderr, "\r    %d bytes...", incnt);
		}
	}

	
	if(format) {		//if format is csv
		
		for (idx=0;idx<16384;idx++) {
			fprintf(fp,"%d;%d;%d;%d\n",cdata[0][idx],cdata[1][idx],cdata[2][idx],cdata[3][idx]);
		}
	}
	else {				//if format is ascii

		(void)fprintf(fp, "Chan1 Chan2 Chan3 Chan4\n");	//header

		for (idx=0;idx<16384;idx++) {
			fprintf(fp,"%d %d %d %d\n",cdata[0][idx],cdata[1][idx],cdata[2][idx],cdata[3][idx]);
		}
	}
	
	
	(void)fprintf(stderr, "* Total bytes transferred: %d ", incnt);
	now_t = gettime();
	(void)fprintf(stderr, "in %lds\n", now_t-last_t);
	if (fclose(fp) != 0)
		err(1, "fclose(%s) failed", fnam);
}

static void
receive_screen(int com, int type)
{
	FILE *fp;
	
	unsigned int llength;			/* RLE */
	unsigned char lastbyte, esc;
	int x, y, outcnt, incnt;
	int r;
	unsigned int bufsize;
	unsigned char c;
	unsigned int i, j;

	switch (type) {
	case 0:
		bufsize=640*480/2;
		rdplanes=maxplanes;
		(void)fprintf(stderr, "  - Receiving a colour screenshot\n");
		break;
	case 1:
		bufsize=640*480/8;   
		rdplanes=maxplanes;
		(void)fprintf(stderr, "  - Receiving a monochrome screenshot\n");
		break;
	default:
		errx(1, "Internal Error:  Unhandled Dumptype in receive_screen(): %d", type);
	}

	/*********************************************************************
	* Read all from stdin/COMx, decompress and fill buffer
	*********************************************************************/
	incnt = outcnt = 0;
	llength = 0;
	lastbyte = esc = 0;
	bufptr= buffer;
	while (((r = rdchar(com)) >= 0) && (bufptr<buffer+bufsize)) {
		c=(unsigned char)r;
		incnt++;
		if (incnt % 100 == 0)
			(void)fprintf(stderr, "\r    %d bytes...", incnt);

		if (esc==2) {
			llength|=(c);					//2nb Byte after ESC is MSB
			for (i=0;i<llength;i++) {
				*bufptr++=lastbyte;
				outcnt++;
			}
			llength=0;
			esc=0; }
		else if (esc==1) {					//2nd escape
			if (c==0xFF) {					//escaped 0xFF
				*bufptr++=c;
				outcnt++;
				lastbyte=c;
				llength=0;
				esc=0;					//Escape-Sequence finished
			} 
			else if (c==0xFE) {		//EOF
				(void)fprintf(stderr, "\nEOF\n");
				bufptr--;						// remove 0x00, not necessary
				break; }
			else {					//1st Byte after ESC is MSB
				llength=c<<8;
				esc++; }
			}
		else if (c != 0xFF) {
			*bufptr++=c;
			outcnt++;
			lastbyte=c;
		} else
			esc=1;
	}
	(void)fprintf(stderr, "\n");
	bufend=bufptr;
	bufptr=buffer;
	
	(void)fprintf(stderr, "* Total bytes transferred: %d(%d) ", outcnt, incnt);
	now_t = gettime();
	(void)fprintf(stderr, "in %lds\n", now_t-last_t);
	last_t = now_t;

	j=0;

	if (type==1){		// BW, is this file really necessary?
		fp = fopen("out.pbm", "w+");
		fprintf(fp,"P1\n640 480\n");
		for (y=0;y<480;y++) {
			for (x=0; x<640/8;x++) {
				c=buffer[y*640/8+x];
				for (j=0;j<8;j++) {			// 8 pixel in a byte
					fprintf(fp,"%d ",((c & (1<<j)) ? 1 : 0));		// ascii output
				}
			}
			fprintf(fp,"\n");
		}
		fclose(fp); // exit(0);
	}
	
	
}

static void
retrieve(int all, int com, int dumptype_req)
{
	int dumptype, r;
	unsigned char c, l;

	if (dumptype_req != -1) {
		(void)fprintf(stderr, "* Triggering Dump...\n");

		c = 0x02;
		if (wrchar(com, c) != 1)
			err(1, "Unable to write to COM port (at 1)");
		#ifdef WIN32
			Sleep(1);
		#else
			(void)sleep(1);
		#endif

		if (dumptype_req == 0)
			c = 0x01;
		else if (dumptype_req == 1)
			c = 0x02;
		else if (dumptype_req == 2)
			c = 0x03;
		else if (dumptype_req == 3)
			c = 0x04;
		else if (dumptype_req != -1)
			errx(1, "Internal Error:  Unhandled Dumptype in retrieve() (at 1): %d", dumptype_req);

		if (wrchar(com, c) != 1)
			err(1, "Unable to write to COM port (at 2)");
	}

	(void)fprintf(stderr, "* Waiting for Screenshot or Trace...\n");

	last_t = gettime();
	l = 0;
	while ((r = rdchar(com)) >= 0) {
		c = (unsigned char)r;
		if ((c == 255) && (l == 'S'))
			break;
		l = c;
	}
	if (r < 0)
		exit(1);

	(void)fprintf(stderr, "* Found Data Start Marker");
	now_t = gettime();
	(void)fprintf(stderr, " after %lds\n", now_t-last_t);
	last_t = now_t;

	/*********************************************************************
	* Next char determines type of data
	*********************************************************************/
	dumptype = (int)rdchar(com);

	switch (dumptype) {
	case 0: /* FALLTHRU */
	case 1:
		receive_screen(com, dumptype);
		combine_screen(!all, dumptype);
		break;
	case 2:
		if (strcasecmp(tracetype, "csv") == 0)
			receive_trace(com, dumptype, "csv", 1);
		else if (strcasecmp(tracetype, "ascii") == 0)
			receive_trace(com, dumptype, "asc", 0);
		else
			errx(1, "Internal Error:  Unhandled Trace Filetype in retrieve() (at 2): %s", tracetype);
		break;
	case 3:
		if (strcasecmp(tracetype, "csv") == 0)
			receive_trace(com, dumptype, "csv", 1);
		else if (strcasecmp(tracetype, "ascii") == 0)
			receive_trace(com, dumptype, "asc", 0);
		else
			errx(1, "Internal Error:  Unhandled Trace Filetype in retrieve() (at 2): %s", tracetype);
		break;
	default:
		errx(1, "Internal Error:  Unhandled Dumptype in retrieve() (at 2): %d", dumptype);
	}
}

static void
combine_screen(int bmp, int dumptype)
{
	#define INVP(colourcomp)	iflag ? (255 - (colourcomp)) : (colourcomp)

	FILE *fp;
	char fnam[256], suf[4];
	unsigned int colour;
	unsigned int x;
	unsigned char j;

	(void)snprintf(suf, sizeof(suf), "ppm");
	if (bmp)
		(void)snprintf(suf, sizeof(suf), "bmp");

	(void)snprintf(fnam, sizeof(fnam), "%s-%.4d.%s", fnamprefix, fnum, suf);
	(void)fprintf(stderr, "* Combining (file: %s)...\n", fnam);
	if ((fp = fopen(fnam, "w+")) == NULL)
		err(1, "fopen(%s, w+) failed", fnam);
	fnum++;

	bufptr=buffer;

	if (bmp) {
		/*
		 * http://en.wikipedia.org/wiki/BMP_file_format
		 */
		/* 0x00 */ (void)fprintf(fp, "BM");
		/* 0x02 */ (void)fprintf(fp, "%c%c%c%c%c%c%c%c", 0, 0, 0, 0, 0, 0, 0, 0);
		/* 0x0A */ (void)fprintf(fp, "%c%c%c%c", 54, 0, 0, 0);		/* 40 (header size) + 0x0A + 4 */
		/* 0x0E */ (void)fprintf(fp, "%c%c%c%c", 40, 0, 0, 0);		/* header size */
		/* 0x12 */ (void)fprintf(fp, "%c%c%c%c", 128, 2, 0, 0);		/* 640px */
		/* 0x16 */ (void)fprintf(fp, "%c%c%c%c", 32, 254, 255, 255);	/* -480px */		//may not work on linux
		/* 0x1A */ (void)fprintf(fp, "%c%c", 1, 0);
		/* 0x1C */ (void)fprintf(fp, "%c%c", 24, 0);			/* bits per pixel */
		/* 0x1E */ (void)fprintf(fp, "%c%c%c%c", 0, 0, 0, 0);		/* no compression */
		/* 0x22 */ (void)fprintf(fp, "%c%c%c%c", 0, 16, 14, 0);		/* bitmap data size: 640*480*3=921600 Bytes */	//
		/* 0x26 */ (void)fprintf(fp, "%c%c%c%c", 19, 11, 0, 0);	//KB	//0,0,0,0
		/* 0x2A */ (void)fprintf(fp, "%c%c%c%c", 19, 11, 0, 0);	//KB	//0,0,0,0
		/* 0x2E */ (void)fprintf(fp, "%c%c%c%c", 0, 0, 0, 0);
		/* 0x32 */ (void)fprintf(fp, "%c%c%c%c", 0, 0, 0, 0);

		if(dumptype==0)	//colour BMP
		{
			for(x=0;x<640*480/2;x++)
			{
				colour=bufptr[x] & 0x0F;
				/* colour order BGR */
				(void)fputc(INVP(planecolours[colour][2]), fp);
				(void)fputc(INVP(planecolours[colour][1]), fp);
				(void)fputc(INVP(planecolours[colour][0]), fp);
	
				colour=bufptr[x] >> 4;
				/* colour order BGR */
				(void)fputc(INVP(planecolours[colour][2]), fp);
				(void)fputc(INVP(planecolours[colour][1]), fp);
				(void)fputc(INVP(planecolours[colour][0]), fp);
			}
		}
		else if (dumptype==1)	//BW BMP
		{
			for(x=0;x<640*480/8;x++)
			{
				for (j=0;j<8;j++)
				{
					if (*bufptr & (1<<j))
					{
						(void)fputc(INVP(planecolours[14][2]), fp);
						(void)fputc(INVP(planecolours[14][1]), fp);
						(void)fputc(INVP(planecolours[14][0]), fp);
					}
					else
					{
						(void)fputc(INVP(planecolours[13][2]), fp);
						(void)fputc(INVP(planecolours[13][1]), fp);
						(void)fputc(INVP(planecolours[13][0]), fp);
					}
				}
				bufptr++;
			}

		}
		
	} else	//PPM
	{
		/*
		 * http://netpbm.sourceforge.net/doc/ppm.html
		 */
		(void)fprintf(fp, "P6\n640 480\n255 ");
		
		if (dumptype==0)		//colour PPM
		{
			for(x=0;x<640*480/2;x++)
			{
				colour=bufptr[x] & 0x0F;
				/* colour order RGB */
				(void)fputc(INVP(planecolours[colour][0]), fp);
				(void)fputc(INVP(planecolours[colour][1]), fp);
				(void)fputc(INVP(planecolours[colour][2]), fp);
	
				colour=bufptr[x] >> 4;
				/* colour order RGB */
				(void)fputc(INVP(planecolours[colour][0]), fp);
				(void)fputc(INVP(planecolours[colour][1]), fp);
				(void)fputc(INVP(planecolours[colour][2]), fp);
			}
		}
		else if (dumptype==1)	//BW PPM
		{
			for(x=0;x<640*480/8;x++)
			{
				for (j=0;j<8;j++)
				{
					if (*bufptr & (1<<j))
					{
						(void)fputc(INVP(planecolours[14][0]), fp);
						(void)fputc(INVP(planecolours[14][1]), fp);
						(void)fputc(INVP(planecolours[14][2]), fp);
					}
					else
					{
						(void)fputc(INVP(planecolours[13][0]), fp);
						(void)fputc(INVP(planecolours[13][1]), fp);
						(void)fputc(INVP(planecolours[13][2]), fp);
					}
				}
				bufptr++;
			}

		}
	}
	
	if (fclose(fp) != 0)
		err(1, "fclose(%s) failed", fnam);

	(void)fprintf(stderr, "* Done\n");
}

static time_t
gettime(void)
{
#ifdef WIN32
	/*
	 * Yes, this is crude.  But I won't implement
	 * a decent API just for Windows.
	 */
	SYSTEMTIME st;

	GetSystemTime(&st);
	return st.wDay * 24*3600 + st.wHour * 3600 + st.wMinute * 60 + st.wSecond;
#else
	return time(NULL);
#endif
}

void	showversion(char *name) {
	fprintf(stderr,"This is %s Rev. %d.%d b\n",name,REVMAJ,REVMIN);
	exit(0);
}

int
main(int argc, char **argv)
{
	int ch, bflag, oneshot, dumptype;
	char *device;
	int com;

	device = NULL;
	bflag = iflag = 0;
	oneshot = 0;
	dumptype = -1;
	fnum = 0;
	(void)snprintf(tracetype, sizeof(tracetype), "csv");
	(void)snprintf(fnamprefix, sizeof(fnamprefix), "screenshot");
	while ((ch = getopt(argc, argv, "c:f:bhsmdrovn:it:")) != -1) {
		switch (ch) {
		case 'c':
#ifdef WIN32
			device = optarg;
			if (*device - '0' < 1 || *device - '0' > 9 || strlen(device) > 1)
				errx(1, "invalid COM port given, use -c n, where 0 < n < 10");
#else
			device = optarg;
#endif
			break;
		case 'f':
			(void)snprintf(fnamprefix, sizeof(fnamprefix), "%s", optarg);
			break;
		case 'v':
			showversion(argv[0]);
			break;
		case 'r':
			dumptype = 3;
			break;
		case 'd':
			dumptype = 2;
			break;
		case 'm':
			dumptype = 1;
			break;
		case 's':
			dumptype = 0;
			break;
		case 'b':
			bflag = 1;
			break;
		case 'o':
			oneshot = 1;
			break;
		case 'n':
			fnum = atoi(optarg);
			if (fnum < 0 || fnum > 9999)
				errx(1, "invalid start of file numbering given, use -n num, where 0 <= num <= 9999");
			break;
		case 'i':
			iflag = 1;
			break;
		case 't':
			(void)snprintf(tracetype, sizeof(tracetype), "%s", optarg);
			if (strcasecmp(tracetype, "csv") != 0 && strcasecmp(tracetype, "ascii") != 0)
				errx(1, "invalid type given, valid: -t ascii and -t csv");
			break;
		case 'h':
		default:
			usage();
			/* NOTREACHED */
		}
	}

	if (dumptype != -1)
		oneshot = 1;

	com = open_serial(device);
//	set_serial(com);

	do {
		retrieve(!bflag, com, dumptype);
	} while (!oneshot);

	return 0;
}

 	  	 
